// Code generated by bpf2go; DO NOT EDIT.
//go:build 386 || amd64

package trace

import (
	"bytes"
	_ "embed"
	"fmt"
	"io"

	"github.com/cilium/ebpf"
)

type bpfEvent struct {
	Meta struct {
		Pc          uint64
		Skb         uint64
		SecondParam uint64
		Mark        uint32
		Netns       uint32
		Ifindex     uint32
		Pid         uint32
		Ifname      [16]uint8
		Pname       [32]uint8
	}
	Tuple struct {
		Saddr struct {
			V4addr uint32
			_      [12]byte
		}
		Daddr struct {
			V4addr uint32
			_      [12]byte
		}
		Sport      uint16
		Dport      uint16
		L3Proto    uint16
		L4Proto    uint8
		TcpFlags   uint8
		PayloadLen uint16
	}
}

// loadBpf returns the embedded CollectionSpec for bpf.
func loadBpf() (*ebpf.CollectionSpec, error) {
	reader := bytes.NewReader(_BpfBytes)
	spec, err := ebpf.LoadCollectionSpecFromReader(reader)
	if err != nil {
		return nil, fmt.Errorf("can't load bpf: %w", err)
	}

	return spec, err
}

// loadBpfObjects loads bpf and converts it into a struct.
//
// The following types are suitable as obj argument:
//
//	*bpfObjects
//	*bpfPrograms
//	*bpfMaps
//
// See ebpf.CollectionSpec.LoadAndAssign documentation for details.
func loadBpfObjects(obj interface{}, opts *ebpf.CollectionOptions) error {
	spec, err := loadBpf()
	if err != nil {
		return err
	}

	return spec.LoadAndAssign(obj, opts)
}

// bpfSpecs contains maps and programs before they are loaded into the kernel.
//
// It can be passed ebpf.CollectionSpec.Assign.
type bpfSpecs struct {
	bpfProgramSpecs
	bpfMapSpecs
}

// bpfSpecs contains programs before they are loaded into the kernel.
//
// It can be passed ebpf.CollectionSpec.Assign.
type bpfProgramSpecs struct {
	KprobeSkb1                   *ebpf.ProgramSpec `ebpf:"kprobe_skb_1"`
	KprobeSkb2                   *ebpf.ProgramSpec `ebpf:"kprobe_skb_2"`
	KprobeSkb3                   *ebpf.ProgramSpec `ebpf:"kprobe_skb_3"`
	KprobeSkb4                   *ebpf.ProgramSpec `ebpf:"kprobe_skb_4"`
	KprobeSkb5                   *ebpf.ProgramSpec `ebpf:"kprobe_skb_5"`
	KprobeSkbLifetimeTermination *ebpf.ProgramSpec `ebpf:"kprobe_skb_lifetime_termination"`
}

// bpfMapSpecs contains maps before they are loaded into the kernel.
//
// It can be passed ebpf.CollectionSpec.Assign.
type bpfMapSpecs struct {
	Events       *ebpf.MapSpec `ebpf:"events"`
	SkbAddresses *ebpf.MapSpec `ebpf:"skb_addresses"`
}

// bpfObjects contains all objects after they have been loaded into the kernel.
//
// It can be passed to loadBpfObjects or ebpf.CollectionSpec.LoadAndAssign.
type bpfObjects struct {
	bpfPrograms
	bpfMaps
}

func (o *bpfObjects) Close() error {
	return _BpfClose(
		&o.bpfPrograms,
		&o.bpfMaps,
	)
}

// bpfMaps contains all maps after they have been loaded into the kernel.
//
// It can be passed to loadBpfObjects or ebpf.CollectionSpec.LoadAndAssign.
type bpfMaps struct {
	Events       *ebpf.Map `ebpf:"events"`
	SkbAddresses *ebpf.Map `ebpf:"skb_addresses"`
}

func (m *bpfMaps) Close() error {
	return _BpfClose(
		m.Events,
		m.SkbAddresses,
	)
}

// bpfPrograms contains all programs after they have been loaded into the kernel.
//
// It can be passed to loadBpfObjects or ebpf.CollectionSpec.LoadAndAssign.
type bpfPrograms struct {
	KprobeSkb1                   *ebpf.Program `ebpf:"kprobe_skb_1"`
	KprobeSkb2                   *ebpf.Program `ebpf:"kprobe_skb_2"`
	KprobeSkb3                   *ebpf.Program `ebpf:"kprobe_skb_3"`
	KprobeSkb4                   *ebpf.Program `ebpf:"kprobe_skb_4"`
	KprobeSkb5                   *ebpf.Program `ebpf:"kprobe_skb_5"`
	KprobeSkbLifetimeTermination *ebpf.Program `ebpf:"kprobe_skb_lifetime_termination"`
}

func (p *bpfPrograms) Close() error {
	return _BpfClose(
		p.KprobeSkb1,
		p.KprobeSkb2,
		p.KprobeSkb3,
		p.KprobeSkb4,
		p.KprobeSkb5,
		p.KprobeSkbLifetimeTermination,
	)
}

func _BpfClose(closers ...io.Closer) error {
	for _, closer := range closers {
		if err := closer.Close(); err != nil {
			return err
		}
	}
	return nil
}

// Do not access this directly.
//
//go:embed bpf_x86_bpfel.o
var _BpfBytes []byte
