// Code generated from java-escape by ANTLR 4.11.1. DO NOT EDIT.

package dae_config // dae_config
import "github.com/antlr/antlr4/runtime/Go/antlr/v4"

// Basedae_configListener is a complete listener for a parse tree produced by dae_configParser.
type Basedae_configListener struct{}

var _ dae_configListener = &Basedae_configListener{}

// VisitTerminal is called when a terminal node is visited.
func (s *Basedae_configListener) VisitTerminal(node antlr.TerminalNode) {}

// VisitErrorNode is called when an error node is visited.
func (s *Basedae_configListener) VisitErrorNode(node antlr.ErrorNode) {}

// EnterEveryRule is called when any rule is entered.
func (s *Basedae_configListener) EnterEveryRule(ctx antlr.ParserRuleContext) {}

// ExitEveryRule is called when any rule is exited.
func (s *Basedae_configListener) ExitEveryRule(ctx antlr.ParserRuleContext) {}

// EnterStart is called when production start is entered.
func (s *Basedae_configListener) EnterStart(ctx *StartContext) {}

// ExitStart is called when production start is exited.
func (s *Basedae_configListener) ExitStart(ctx *StartContext) {}

// EnterBare_literal is called when production bare_literal is entered.
func (s *Basedae_configListener) EnterBare_literal(ctx *Bare_literalContext) {}

// ExitBare_literal is called when production bare_literal is exited.
func (s *Basedae_configListener) ExitBare_literal(ctx *Bare_literalContext) {}

// EnterQuote_literal is called when production quote_literal is entered.
func (s *Basedae_configListener) EnterQuote_literal(ctx *Quote_literalContext) {}

// ExitQuote_literal is called when production quote_literal is exited.
func (s *Basedae_configListener) ExitQuote_literal(ctx *Quote_literalContext) {}

// EnterLiteral is called when production literal is entered.
func (s *Basedae_configListener) EnterLiteral(ctx *LiteralContext) {}

// ExitLiteral is called when production literal is exited.
func (s *Basedae_configListener) ExitLiteral(ctx *LiteralContext) {}

// EnterLiteralExpression is called when production literalExpression is entered.
func (s *Basedae_configListener) EnterLiteralExpression(ctx *LiteralExpressionContext) {}

// ExitLiteralExpression is called when production literalExpression is exited.
func (s *Basedae_configListener) ExitLiteralExpression(ctx *LiteralExpressionContext) {}

// EnterInput is called when production input is entered.
func (s *Basedae_configListener) EnterInput(ctx *InputContext) {}

// ExitInput is called when production input is exited.
func (s *Basedae_configListener) ExitInput(ctx *InputContext) {}

// EnterProgramStructureBlcok is called when production programStructureBlcok is entered.
func (s *Basedae_configListener) EnterProgramStructureBlcok(ctx *ProgramStructureBlcokContext) {}

// ExitProgramStructureBlcok is called when production programStructureBlcok is exited.
func (s *Basedae_configListener) ExitProgramStructureBlcok(ctx *ProgramStructureBlcokContext) {}

// EnterExpression is called when production expression is entered.
func (s *Basedae_configListener) EnterExpression(ctx *ExpressionContext) {}

// ExitExpression is called when production expression is exited.
func (s *Basedae_configListener) ExitExpression(ctx *ExpressionContext) {}

// EnterDeclaration is called when production declaration is entered.
func (s *Basedae_configListener) EnterDeclaration(ctx *DeclarationContext) {}

// ExitDeclaration is called when production declaration is exited.
func (s *Basedae_configListener) ExitDeclaration(ctx *DeclarationContext) {}

// EnterOptAnnotation is called when production optAnnotation is entered.
func (s *Basedae_configListener) EnterOptAnnotation(ctx *OptAnnotationContext) {}

// ExitOptAnnotation is called when production optAnnotation is exited.
func (s *Basedae_configListener) ExitOptAnnotation(ctx *OptAnnotationContext) {}

// EnterFunctionPrototype is called when production functionPrototype is entered.
func (s *Basedae_configListener) EnterFunctionPrototype(ctx *FunctionPrototypeContext) {}

// ExitFunctionPrototype is called when production functionPrototype is exited.
func (s *Basedae_configListener) ExitFunctionPrototype(ctx *FunctionPrototypeContext) {}

// EnterOptParameterList is called when production optParameterList is entered.
func (s *Basedae_configListener) EnterOptParameterList(ctx *OptParameterListContext) {}

// ExitOptParameterList is called when production optParameterList is exited.
func (s *Basedae_configListener) ExitOptParameterList(ctx *OptParameterListContext) {}

// EnterNonEmptyParameterList is called when production nonEmptyParameterList is entered.
func (s *Basedae_configListener) EnterNonEmptyParameterList(ctx *NonEmptyParameterListContext) {}

// ExitNonEmptyParameterList is called when production nonEmptyParameterList is exited.
func (s *Basedae_configListener) ExitNonEmptyParameterList(ctx *NonEmptyParameterListContext) {}

// EnterParameter is called when production parameter is entered.
func (s *Basedae_configListener) EnterParameter(ctx *ParameterContext) {}

// ExitParameter is called when production parameter is exited.
func (s *Basedae_configListener) ExitParameter(ctx *ParameterContext) {}

// EnterRoutingRule is called when production routingRule is entered.
func (s *Basedae_configListener) EnterRoutingRule(ctx *RoutingRuleContext) {}

// ExitRoutingRule is called when production routingRule is exited.
func (s *Basedae_configListener) ExitRoutingRule(ctx *RoutingRuleContext) {}

// EnterOutboundExpr is called when production outboundExpr is entered.
func (s *Basedae_configListener) EnterOutboundExpr(ctx *OutboundExprContext) {}

// ExitOutboundExpr is called when production outboundExpr is exited.
func (s *Basedae_configListener) ExitOutboundExpr(ctx *OutboundExprContext) {}

// EnterFunctionPrototypeExpression is called when production functionPrototypeExpression is entered.
func (s *Basedae_configListener) EnterFunctionPrototypeExpression(ctx *FunctionPrototypeExpressionContext) {
}

// ExitFunctionPrototypeExpression is called when production functionPrototypeExpression is exited.
func (s *Basedae_configListener) ExitFunctionPrototypeExpression(ctx *FunctionPrototypeExpressionContext) {
}

// EnterRoutingRuleOrDeclarationOrLiteralOrExpressionList is called when production routingRuleOrDeclarationOrLiteralOrExpressionList is entered.
func (s *Basedae_configListener) EnterRoutingRuleOrDeclarationOrLiteralOrExpressionList(ctx *RoutingRuleOrDeclarationOrLiteralOrExpressionListContext) {
}

// ExitRoutingRuleOrDeclarationOrLiteralOrExpressionList is called when production routingRuleOrDeclarationOrLiteralOrExpressionList is exited.
func (s *Basedae_configListener) ExitRoutingRuleOrDeclarationOrLiteralOrExpressionList(ctx *RoutingRuleOrDeclarationOrLiteralOrExpressionListContext) {
}

// EnterRoutingRuleList is called when production routingRuleList is entered.
func (s *Basedae_configListener) EnterRoutingRuleList(ctx *RoutingRuleListContext) {}

// ExitRoutingRuleList is called when production routingRuleList is exited.
func (s *Basedae_configListener) ExitRoutingRuleList(ctx *RoutingRuleListContext) {}
